;; Simplest approach to planarity testing - one calculation, no attempt
;; to optimize placement of communicating processors.  Interaction processors
;; are placed in consecutive cube addresses, edge representatives also in
;; consecutive cube addresses.  We set up a representation for
;; a complete graph on n vertices, but disable edges not present.
;; This implements exactly the algorithm given by Takefuji and Lee, which
;; is probably not what they actually ran.

(defvar program 'planar1)

(defconstant NS 10)  ;; Size of vertex number (bits)

(*proclaim '(type fixnum nv nte))

(defvar nv  4)         ;; Number of vertices in graph
(defvar nte 2)         ;; Number of non-trivial edges in complete graph

(*proclaim '(type Boolean-pvar edgep edgsg ntrve edgex intrp intsg prsnt))

(*defvar edgep nil!!)      ;; This processor represents an edge
(*defvar edgsg nil!!)      ;; Segmentation pvar for edges
(*defvar ntrve nil!!)      ;; This edge is non-trivial
(*defvar intrp nil!!)      ;; This processor manages an interaction
(*defvar intsg nil!!)      ;; Segmentation pvar for intrp's
(*defvar prsnt t!!)        ;; When edgep, this edge is present in the graph

(*proclaim '(type (field-pvar NS) edga edgb ntxa ntxb ntra ntrb ntrc ntrc))

(*defvar edga (!! 0))      ;; When edgep, edge is [edga, edgb]
(*defvar edgb (!! 0))

(*defvar ntra (!! 0))      ;; When intrp, interaction is between
(*defvar ntrb (!! 0))      ;;  [ntra, ntrb] and [ntrc, ntrd],
(*defvar ntrc (!! 0))      ;; whence
(*defvar ntrd (!! 0))      ;;  0 <= ntra < ntrc < ntrb < ntrd < nv

(*proclaim '(type (field-pvar 21) self rnvz srce rpab rpcd))

(*defvar self (self-address!!)) ;; For convenience
(*defvar rnvz (!! 0))           ;; For rendezvous calculations
(*defvar rpab (!! 0))           ;; When intrp, rep of [a,b]
(*defvar rpcd (!! 0))           ;; When intrp, rep of [c,d]

(*proclaim '(type (signed-pvar 24) cntx))

(*defvar cntx (!! 0))       ;; For enumeration, etc.

(defun setup (&optional (xs 128) (ys 64))
 (*cold-boot :initial-dimensions (list xs ys))
)

;; Calculate cube address of edge representative for edge [xi, xj]

(*defun ep(xi xj)
  (declare (type (field-pvar NS) xi xj))
  (+!! (ash!! (*!! xi (-!! (!! (1- (ash nv 1))) xi)) (!! -1))
       (-!! xj xi (!! 1)))
)

(defun mk-graph ()
 ;; Make temporary vertices and begin base edge construction
 (*set edgsg nil!!)  (*set edga (!! 0))
 (*when (<!! self (!! nv)) ; Temporary vertex rep.
   (*set cntx (-!! (!! nv) (1+!! self)))     ; Number of edges originating here
   (*set rnvz (-!! (scan!! cntx '+!!) cntx)) ; First edge address
   (*pset :no-collisions self edga rnvz :notify edgsg)
                                 ; Originating vertex to first edge
 )
 ;; Fill in base edge representations
 (*set edga (scan!! edga 'copy!! :segment-pvar edgsg))  ; Origins
 (*when (<!! edga (!! (1- nv)))   ; Real edge rep.
   (*set edgep t!!)
   (*set edgb (+!! edga (scan!! (!! 1) '+!! :segment-pvar edgsg))) ; Dest.
   (*set ntrve (not!! (or!! (=!! edgb (1+!! edga))   ; Non-trivial edges
                            (and!! (zerop!! edga) (=!! edgb (!! (1- nv)))))))
 )
 ;; Calculate number of interactions for non-trivial edges and
 ;; start construction of interaction processors
 (*when ntrve
   (*set cntx (*!! (-!! edgb edga (!! 1)) (-!! (!! (1- nv)) edgb))) ;interact's
   (*when (plusp!! cntx)      ; Actually responsible for interactions
     (*set rnvz (-!! (scan!! cntx '+!!) cntx))   ; First intrp address
     (*pset :no-collisions edga ntra rnvz :notify intsg) ; a to first intrp
     (*pset :no-collisions edgb ntrb rnvz)               ; and b
   )
 )
 ;; Form interaction processors
 (*set cntx ( 1-!! (scan!! (!! 1) '+!! :segment-pvar intsg)))
                                               ;;Position in intsg
 (*set ntra (scan!! ntra 'copy!! :segment-pvar intsg))
 (*set ntrb (scan!! ntrb 'copy!! :segment-pvar intsg))
 (*let ((isz (1-!! (-!! ntrb ntra))) (irt (-!! (!! (1- nv)) ntrb)))
       (declare (type (signed-pvar (1+ NS)) isz irt))
   (*when (<!! cntx (*!! isz irt))
      (*set intrp t!!)  ;; Actual interaction processor
      (*set ntrc (+!! ntra (truncate!! cntx irt) (!! 1)))
      (*set ntrd (+!! ntrb (rem!! cntx irt) (!! 1)))
      (*set rpab (ep ntra ntrb))  (*set rpcd (ep ntrc ntrd))
   )
 )
)

;; Neural net state and parameters

(*proclaim '(type Boolean-pvar Vup Vdn acte actr))

(*defvar Vup  nil!!)   ;; Up edge
(*defvar Vdn  nil!!)   ;; Down edge
(*defvar acte nil!!)   ;; edgep and ntrve and prsnt (active edge rep)
(*defvar actr nil!!)   ;; actvive interaction

(*proclaim '(type single-float-pvar Uup Udn Du PD))

(*defvar Uup (!! 0.0))  ;; Up activation
(*defvar Udn (!! 0.0))  ;; Down activation
(*defvar Du  (!! 0.0))  ;; Delta U (for both up and down in turn)
(*defvar PD  (!! 0.0))  ;; Up-down contribution to Du

(*proclaim '(type single-float A B DT))

(defvar A  2.0)       ;; Coefficient for up-down conflict
(defvar B  1.0)       ;; Coefficient for edge conflicts
(defvar DT 1.0E-5)    ;; Delta-t

(defvar cycles 0)     ;; Counts iterations of nueral algorithm
(defvar legal nil)    ;; Used to report conflicts, T iff last configuration
                      ;; is legal
(defvar more nil)     ;; Used to report potential for more edges, true iff
                      ;; Du > 0 somewhere

;; Random float!! between lo and hi

(*defun rf!! (lo hi)
 (+!! (*!! (random!! (!! 1000000)) (!! (/ (- hi lo) 1E6))) (!! lo))
)

;; Start neural-net calculation, after T&L, parameters are lower, upper
;; bounds for initial activations.  Handles trivial edges.

(defun start (&optional (Ul (- DT)) (Uh 0.0))
 (setq cycles 0)
 (*set acte (and!! edgep ntrve prsnt))
 (*set actr (and!! (pref!! acte rpab) (pref!! acte rpcd)))
 (*when acte
   (*set Uup (rf!! Ul Uh))  (*set Udn (rf!! Ul Uh))
 )
 (*set Vup nil!!)  (*set Vdn nil!!)
 (*when (and!! edgep (not!! ntrve) prsnt)  ;; Trivial edges
   (*set Vup (evenp!! edga)) (*set Vdn (not!! Vup))
 )
)

;; Perform one cycle of algorithm, leaves info for convergence tests

(defun cycle ()
 ;; Calculate V's, PD
 (*when acte
   (*set Vup (plusp!! Uup)) (*set Vdn (plusp!! Udn))
   (*set cntx (!! 0))  ;; To count interactions
   (*let ((V2 (and!! Vup Vdn)))
         (declare (type Boolean-pvar V2))
       (setq legal (not (*or V2)))
       (*set PD (cond!! (V2 (!! (- A)))
                        ((xor!! Vup Vdn) (!! 0.0))
                        (t!! (!! A))))
   ) ; end *let
 ) ; end *when acte
 ;; Calculate Up interactions
 (*when actr
    (*if (pref!! Vup rpab) (*pset :add (!! 1) cntx rpcd))
    (*if (pref!! Vup rpcd) (*pset :add (!! 1) cntx rpab))
 ) ; end *when actr
 ;; Calculate new Up activations
 (*when acte
    (*set Du (-!! PD (*!! (!! B) cntx)))
    (setq more (*or (plusp!! Du)))
    (*set Uup (+!! Uup (*!! Du (!! DT))))
    (setq legal (and legal (*and (or!! (not!! Vup) (zerop!! cntx)))))
    (*set cntx (!! 0)) ;; For Down interactions
 ) ; end *when acte
 ;; Calculate Down interactions
 (*when actr
    (*if (pref!! Vdn rpab) (*pset :add (!! 1) cntx rpcd))
    (*if (pref!! Vdn rpcd) (*pset :add (!! 1) cntx rpab))
 ) ; end *when actr
 ;; Calculate new Down activations
 (*when acte
    (*set Du (-!! PD (*!! (!! B) cntx)))
    (setq more (or more (*or (plusp!! Du))))
    (*set Udn (+!! Udn (*!! Du (!! DT))))
    (setq legal (and legal (*and (or!! (not!! Vdn) (zerop!! cntx)))))
 ) ; end *when acte
 (incf cycles)
 (list 'Legal legal 'More more 'Cycles cycles)
)
